/* --------------------------------------------------
 * <RASP-LC-LAN>
 * Reference : rasp.c
 * -------------------------------------------------- */

#include "rasp_lcw.h"
#include <limits.h>

#define PERIODSIZE 928

typedef int rasp_pcm_format_t;

typedef struct {
	WS_HANDLE_t * device;
	wios_buffer_t buffer;
	char * ip_addr;
} rasp_priv_t;



/* STATIC FUNCTION PROTOTYPE */
/* -------------------------------------------------- */

static int priv_init(wios_file_t * ft);
static int device_read(wios_file_t * ft, wios_buffer_t * buf, int len);
static char * rasp_buffer_malloc(wios_file_t * ft, wios_buffer_t * buf, int first_nframes);

/* HANDLER FUNCTION */
/* -------------------------------------------------- */
static int device_init(wios_file_t * ft)
{
	rasp_priv_t    * priv = ft->priv;
	wios_fmtinfo_t * fmt = &ft->fmtinfo;
	int header_flag = 0; /* RASPヘッダをなしに設定する */
	int status;

	/* デバイスハンドル作成 */
	if ((priv->device = WsLcHandleOpen(priv->ip_addr)) == NULL)
		WIOS_ERROR("cannot allocate hardware parameter strucutre", NULL);

	/* デバイスハンドルを設定 */
	ft->dp = priv->device;

	/* サンプリングレートを設定 */
	/* 量子化ビット数を設定 */
	/* チャンネル数を設定 */
	/* DAとADのゲインを設定 */
	/* A/Dデータにヘッダ情報を付加しないように設定 */

	return WIOS_SUCCESS;

error:
	return WIOS_EOF;
}

static wios_file_t * open_read(
	const char             * path,
	const char             * dev,
	const wios_execinfo_t  * info)
{
	return open_read_(path, dev, info, priv_init);
}

static wios_file_t * open_write(
		const char             * path,
		const char             * dev,
		const wios_execinfo_t  * info,
		const wios_fmtinfo_t   * fmt)
{
	/* RASPの場合は特別にperiodsizeを設定 */
	wios_execinfo_t * rasp_info = (wios_execinfo_t *)info;
	rasp_info-> periodsize = PERIODSIZE;
	return open_write_(path, dev, info, fmt, priv_init);
}

/* --------------------------------------------------
 * RASP24
 * 再生準備
 *
 * RASPは最大1MB分のデータをバッファ可能。
 * リファレンスにはD/A開始前に32kB以上のデータがないとD/Aは開始しないとあるが、
 * 自分の環境では64kB程度のデータを書きこまないとD/Aが開始されなかったため、
 * 64kB以上書き込んでから再生するようにしてある。
 * -------------------------------------------------- */
static void * playback(void * arg)
{
	return NULL;
}

/* --------------------------------------------------
 * RASP24
 * キャプチャーに関しての注意
 *  
 * periodsizeはAPIに928[frame]で固定されている。
 * APIはこの値を仮定して実装されているため
 * periodsizeは928[frame]に固定してある。
 * -------------------------------------------------- */
static void * capture(void * arg)
{
	/* 録音準備が終わるまで呼び出し元をロック */
	pthread_mutex_lock(&ready_mutex);
	wios_file_t * ft = (wios_file_t *)arg;
	rasp_priv_t * priv = ft->priv;
	wios_buffer_t * buf = &priv->buffer;
	wios_signal_t * sig = &ft->signal;

	int frames;
	int done = 0;
	int len = PERIODSIZE;
	int req_nframes;

	int written = 0;

	int status;

	if (wios_mode == wios_sync && !use_2dev_option_flag) {
		if (ft->signal.execinfo.sec == WIOS_ALL_SEC)
			req_nframes = ftda->signal.all_frames;
		else
			req_nframes = ft->fmtinfo.rate * ft->signal.execinfo.sec;
	} else {
		req_nframes = ft->fmtinfo.rate * ft->signal.execinfo.sec;
	}

	if (rasp_buffer_malloc(ft, buf, ft->signal.buf_size) == NULL)
		goto error;

	/* 録音準備完了をメインスレッドに知らせる */
	pthread_cond_signal(&ready_cond);
	pthread_mutex_unlock(&ready_mutex);

	if (wios_mode == wios_sync) {
		/* メインスレッドで同期を開始させるまで待機 */
		pthread_mutex_lock(&sync_mutex);
		while (!sync_start_flag)
			pthread_cond_wait(&sync_cond, &sync_mutex);
		pthread_mutex_unlock(&sync_mutex);
	}

	/* 以下同期スタート */
	if (print_mesg_flag)
		fprintf(stdout, "Start capture (%s)\n", ft->file_name);
  if (print_time_flag)
    fprintf(stdout, "Capture_start_Time:%lf\n", wios_get_time());

	if ((status = WsLcADStart(ft->dp)) != 1)
		WIOS_ERROR("cannot start playback (status:0x%08x)", status);


	/* 録音ループ */
	while (1) {
		/* SIGINTを受け取っていたらブレーク */
		if (loop_break_flag) break;

		frames = device_read(ft, buf, len);

		if (wios_mode != wios_rec || ft->signal.execinfo.sec != WIOS_ALL_SEC) {
			/* 余分なフレームを書きこまないようにする */
			if (frames > req_nframes-done) {
				if (ft->fmtinfo.format == WIOS_PCM_FORMAT_S24)
					ft->success_byte_cnt -= (frames-(req_nframes-done))*3*ft->fmtinfo.channels;
				else 
					ft->success_byte_cnt -= (frames-(req_nframes-done))*ft->signal.byte_per_item*ft->fmtinfo.channels;
				frames = req_nframes-done;
			}
		}

		lsf_file_write(ft, buf->fl, frames);

		done += frames;

		/* 録音モードで秒数が指定されなかった場合はブレークしない */
		if (wios_mode != wios_rec || ft->signal.execinfo.sec != WIOS_ALL_SEC)
			if (done >= req_nframes) break;
	}

	WsLcADStop(ft->dp);

	if (print_mesg_flag)
		fprintf(stdout, "End capture (%s)\n", ft->file_name);

	wios_buffer_free(buf);

	return NULL;

error:
	if (pthread_mutex_destroy(&ready_mutex) == EBUSY)
		pthread_mutex_unlock(&ready_mutex);

	wios_buffer_free(buf);

	return NULL;
}

static void finish(wios_file_t * ft)
{
	rasp_priv_t * priv = ft->priv;
	if (ft) {
		if (ft->fp) {
			sf_close(ft->fp);
			ft->fp = NULL;
		}
		if (ft->dp) {
			WsLcHandleClose(ft->dp);
			ft->dp = NULL;
		}
		if (priv) {
			free(priv);
			priv = NULL;
		}
		if (ft) {
			free(ft);
			ft = NULL;
		}
	}
}



/* STATIC FUNCTION */
/* -------------------------------------------------- */

static int priv_init(wios_file_t * ft)
{
	rasp_priv_t * priv;

	if ((ft->priv = malloc(sizeof(rasp_priv_t))) == NULL)
		goto error;
	priv = ft->priv;

	/* RASP仕様時はデバイス名をIPアドレスをとする */
	priv->ip_addr = (char *)ft->dev_name;

	priv->buffer.fl = NULL;
	priv->buffer.dv = NULL;

	return WIOS_SUCCESS;

error:
	return WIOS_EOF;
}


static int device_read(wios_file_t * ft, wios_buffer_t * buf, int nframes)
{
	int i, err;
	int n = ft->signal.byte_per_item*ft->fmtinfo.channels;

	/* nframes*n[byte]のデータをデバイスから読み込む */
	err = WsLcDataRead(ft->dp, buf->dv, nframes*n);

	if (err == nframes*n) {
		ft->success_byte_cnt += err > 0 ? (err*3)/4 : 0;
	} else {
		ft->success_byte_cnt += err > 0 ? (err*3)/4 : 0;
		ft->missing_byte_cnt += (nframes*n-err) > 0 ? (nframes*n-err*3)/4 : 0;
	}

	int24_t * buf1 = (int24_t *)buf->dv;
	for (i = 0; i < nframes*ft->fmtinfo.channels; i++) {
		buf->fl[i] = WIOS_CONVERT_24_TO_32BIT(buf1[i]);
		AD_CLIP_WARN(WIOS_24BIT_CLIP_COUNT(buf1[i], ft->clips));
		WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
	}

	return err/n;
}

char * rasp_buffer_malloc(wios_file_t * ft, wios_buffer_t * buf, int first_nframes)
{
	return wios_buffer_malloc(ft, buf, first_nframes);
}



/* GLOBAL FUNCTION */
/* -------------------------------------------------- */

wios_format_handler_t const * wios_set_rasp_lcw_format_handler(void)
{
	static wios_format_handler_t const handler = {
		device_init,
		open_read, open_write,
		playback, capture,
		finish
	};

	return &handler;
}
